#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <asm/pgtable.h>
#include <asm/pgtable-hwdef.h> // for ARMv7-specific page table defs
#include <asm/memory.h> // for virtual-to-physical address translation
#include <asm/mach/map.h> // for memory type defs
#include <asm/pgalloc.h>
#include <asm/tlbflush.h>
#include <asm/cacheflush.h>
#include <linux/mman.h>
#include <linux/mm.h>
#include <linux/sched.h> // has reference to extern'd struct mm_struct init_mm
#include <linux/vmalloc.h>
#include <linux/hrtimer.h>
#include <linux/proc_fs.h>
#include <linux/irqdomain.h>
#include <linux/gpio.h>
#include <linux/clk.h>
#include <linux/device.h>
#include <linux/i2c.h>
#include <video/omapdss.h>
//#include <linux/platform_data/gpio-omap.h>

#include <linux/remap_pgtbl.h> // for hacking the kernel-space page fault handler

#define DEBUG_LEVEL 0

#define PRINTK_LEVEL(x, fmt, args...)  \
	if (DEBUG_LEVEL >= x) printk(fmt, ##args);

#define PRINTK0(fmt, args...)  PRINTK_LEVEL(0, fmt, ##args);
#define PRINTK1(fmt, args...)  PRINTK_LEVEL(1, fmt, ##args);
#define PRINTK2(fmt, args...)  PRINTK_LEVEL(2, fmt, ##args);
#define PRINTK3(fmt, args...)  PRINTK_LEVEL(3, fmt, ##args);



#define FSR_WRITE		(1 << 11) // from mm/fault.c
#define VM_BAD_ACCESS	0x020000;


struct remapped_section {
	struct list_head list;

	pmd_t *pmdp;
	pmd_t original_pmd[2];
	pmd_t new_pmd[2];
};

struct removed_base {
	struct list_head node;
	void __iomem *base;		/* virtual */
	//no need to record these. Reading permissions should be light weight.
//	bool read; //read permission
//	bool write; //write permission
	bool accessed;
//avoid continuous reports
	int catched_times;
	unsigned long long abs_catched_times;
	pte_t *ptep;
	unsigned long pfn;
	void __iomem *clkctrl;
	struct omap_hwmod *oh;
	int power_on_off;
};

int pages2sections(unsigned long vaddr);
int sections2pages(unsigned long vaddr);
int page_access_fix(unsigned long vaddr, unsigned int fsr, struct mm_struct *mm);

struct gpio_regs {
	u32 irqenable1;
	u32 irqenable2;
	u32 wake_en;
	u32 ctrl;
	u32 oe;
	u32 leveldetect0;
	u32 leveldetect1;
	u32 risingdetect;
	u32 fallingdetect;
	u32 dataout;
	u32 debounce;
	u32 debounce_en;
};

//3.12rc5
//struct gpio_bank {
//	struct list_head node;
//	void __iomem *base;
//	u16 irq;
//	struct irq_domain *domain;
//	u32 non_wakeup_gpios;
//	u32 enabled_non_wakeup_gpios;
//	struct gpio_regs context;
//	u32 saved_datain;
//	u32 level_mask;
//	u32 toggle_mask;
//	spinlock_t lock;
//	struct gpio_chip chip;
//	struct clk *dbck;
//	u32 mod_usage;
//	u32 irq_usage;
//	u32 dbck_enable_mask;
//	bool dbck_enabled;
//	struct device *dev;
//	bool is_mpuio;
//	bool dbck_flag;
//	bool loses_context;
//	bool context_valid;
//	int stride;
//	u32 width;
//	int context_loss_count;
//	int power_mode;
//	bool workaround_enabled;
//
//	void (*set_dataout)(struct gpio_bank *bank, int gpio, int enable);
//	int (*get_context_loss_count)(struct device *dev);
//
//	struct omap_gpio_reg_offs *regs;
//};

struct gpio_bank {
	//CHAO:
	struct list_head node;

	unsigned long pbase;
	void __iomem *base;
	u16 irq;
	u16 virtual_irq_start;
	int method;
	u32 suspend_wakeup;
//#if defined(CONFIG_ARCH_OMAP16XX) || defined(CONFIG_ARCH_OMAP2PLUS)
	u32 saved_wakeup;
//#endif
	u32 non_wakeup_gpios;
	u32 enabled_non_wakeup_gpios;

	u32 saved_datain;
	u32 saved_fallingdetect;
	u32 saved_risingdetect;
	u32 level_mask;
	u32 toggle_mask;
	spinlock_t lock;
	struct gpio_chip chip;
	struct clk *dbck;
	u32 mod_usage;
	u32 dbck_enable_mask;
	struct device *dev;
	bool dbck_flag;
	int stride;
	u32 width;

	void (*set_dataout)(struct gpio_bank *bank, int gpio, int enable);

	struct omap_gpio_reg_offs *regs;
};

struct omap_i2c_dev {
	//CHAO:
	struct list_head node;

	struct device		*dev;
	void __iomem		*base;		/* virtual */
	int			irq;
	int			reg_shift;      /* bit shift for I2C register addresses */
	struct completion	cmd_complete;
	struct resource		*ioarea;
	u32			latency;	/* maximum mpu wkup latency */
	void			(*set_mpu_wkup_lat)(struct device *dev,
						    long latency);
	u32			speed;		/* Speed of bus in Khz */
	u16			cmd_err;
	u8			*buf;
	u8			*regs;
	size_t			buf_len;
	struct i2c_adapter	adapter;
	u8			fifo_size;	/* use as flag and value
						 * fifo_size==0 implies no fifo
						 * if set, should be trsh+1
						 */
	u8			rev;
	unsigned		b_hw:1;		/* bad h/w fixes */
	u16			iestate;	/* Saved interrupt register */
	u16			pscstate;
	u16			scllstate;
	u16			sclhstate;
	u16			bufstate;
	u16			syscstate;
	u16			westate;
	u16			errata;
};

struct omap_hsmmc_next {
	unsigned int	dma_len;
	s32		cookie;
};

struct omap_hsmmc_host {
	struct list_head node;

	struct	device		*dev;
	struct	mmc_host	*mmc;
	struct	mmc_request	*mrq;
	struct	mmc_command	*cmd;
	struct	mmc_data	*data;
	struct	clk		*fclk;
	struct	clk		*dbclk;
	/*
	 * vcc == configured supply
	 * vcc_aux == optional
	 *   -	MMC1, supply for DAT4..DAT7
	 *   -	MMC2/MMC2, external level shifter voltage supply, for
	 *	chip (SDIO, eMMC, etc) or transceiver (MMC2 only)
	 */
	struct	regulator	*vcc;
	struct	regulator	*vcc_aux;
	struct	work_struct	mmc_carddetect_work;
	void	__iomem		*base;
	resource_size_t		mapbase;
	spinlock_t		irq_lock; /* Prevent races with irq handler */
	unsigned int		id;
	unsigned int		dma_len;
	unsigned int		dma_sg_idx;
	unsigned char		bus_mode;
	unsigned char		power_mode;
	u32			*buffer;
	u32			bytesleft;
	int			suspended;
	int			irq;
	int			use_dma, dma_ch;
	int			dma_line_tx, dma_line_rx;
	int			slot_id;
	int			got_dbclk;
	int			response_busy;
	int			context_loss;
	int			dpm_state;
	int			vdd;
	int			protect_card;
	int			reqs_blocked;
	int			use_reg;
	int			req_in_progress;
	struct omap_hsmmc_next	next_data;

	struct	omap_mmc_platform_data	*pdata;
};

#define MAX_DSS_OVERLAYS	4
#define DISPC_MAX_NR_ISRS		8
#define SZ_4K				0x00001000
#define DISPC_SZ_REGS			SZ_4K

struct omap_dispc_isr_data {
	omap_dispc_isr_t	isr;
	void			*arg;
	u32			mask;
};

struct struct_dispc{
	struct platform_device *pdev;
	void __iomem    *base;

	int		ctx_loss_cnt;
	struct mutex	runtime_lock;
	int		runtime_count;

	int irq;
	struct clk *dss_clk;

	u32	fifo_size[MAX_DSS_OVERLAYS];

	u32	channel_irq[3]; /* Max channels hardcoded to 3*/

	spinlock_t irq_lock;
	u32 irq_error_mask;
	struct omap_dispc_isr_data registered_isr[DISPC_MAX_NR_ISRS];
	u32 error_irqs;
	struct work_struct error_work;

	bool		ctx_valid;
	u32		ctx[DISPC_SZ_REGS / sizeof(u32)];
};
