/*
 * Special thanks to Kevin Boos for function prototype of page_access_fix()
 * and remove_permissions().
 * Special thanks to Felix Xiaozhu Lin for prototype code that harnesses
 * the performance counter.
 */

#include "local.h"
#include <../mm/mm.h>
#include <linux/chao-pgtbl.h>
#include <plat/omap_device.h>
#include <plat/mmc.h>
#include <cminst44xx.h>
#include <clockdomain.h>
#include <cm-regbits-44xx.h>
#include <linux/delay.h>
#include <linux/gfp.h>

#include <omap_hwmod.h>
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Chao Xu, Kevin Boos");
MODULE_DESCRIPTION("Central PM Agent");

#define TIME_FLUSH 0
#define TIMING 0
#define FINE_FLUSH 1

#define hw_pte(pte) (((unsigned long)pte) + 2048)

#define CMD_PROCFS_MAX_SIZE		1024
#define BUG_PROCFS_MAX_SIZE		1024*1024
#define CMD_PROCFS_NAME 		"cmd-remap"
#define BUG_PROCFS_NAME		"lost-pm-opp"
#define TIME_PROCFS_NAME		"time"
#define MONITOR_PERIOD		50
//report every 10s
#define REPORT_PERIOD 1
//#define REPORT_PERIOD		(10000/MONITOR_PERIOD)

int is_sdio_on=0;

static struct proc_dir_entry *cmd_procfs;
static struct proc_dir_entry *bug_procfs;
static struct proc_dir_entry *time_procfs;
static char cmd_buffer[CMD_PROCFS_MAX_SIZE];
static unsigned long cmd_buffer_size = 0;
static char bug_buffer[BUG_PROCFS_MAX_SIZE];
static unsigned long page_fix_access_count = 0;
static unsigned long untrapped_count = 0;
/*
 * Global variables used to timing for how much time a module is disabled
 */
unsigned long long start_time = 0;
unsigned long long central_elapsed_time = 0;
unsigned long long start_time_copy=10;
unsigned long long end_time = 0;

/*
 * A list of all module controlled by Central PM Agent
 */
struct list_head removed_base_list;

void setup_gpio_base_protection(int);
void setup_i2c_base_protection(int);
void setup_mmc_base_protection(int);
void test_setup_gpio_base_protection(void);
void test_setup_i2c_base_protection(void);
void remove_permissions(unsigned long, bool, struct removed_base *);
void test_remove_permissions(unsigned long, bool, struct removed_base *);
void measure_page_fault(void);
void test_get_cycle(void);
void setup_dss_base_protection(void);

struct remap_pgtbl_ops ops = { .page_access_fix = page_access_fix };

static struct timer_list my_timer;

u32 read_clkctrl_from_oh(struct omap_hwmod *oh) {
	return omap4_cminst_read_inst_reg(oh->clkdm->prcm_partition,
			oh->clkdm->cm_inst, oh->prcm.omap4.clkctrl_offs);
}

/*
 * Performance counters are used to measure the latency caused by page-fault
 * etc.
 */
void init_perfcounters(int32_t do_reset, int32_t enable_divider) {
	// in general enable all counters (including cycle counter)
	int32_t value = 1;

	// perform reset:
	if (do_reset) {
		value |= 2;     // reset all counters to zero.
		value |= 4;     // reset cycle counter to zero.
	}

	if (enable_divider)
		value |= 8;     // enable "by 64" divider for CCNT.

	value |= 16;

	// program the performance-counter control-register:
	asm volatile ("MCR p15, 0, %0, c9, c12, 0\t\n" :: "r"(value));

	// enable all counters:
	asm volatile ("MCR p15, 0, %0, c9, c12, 1\t\n" :: "r"(0x8000000f));

	// clear overflows:
	asm volatile ("MCR p15, 0, %0, c9, c12, 3\t\n" :: "r"(0x8000000f));
}

/* ~60-100 cycles on omap4460 */
static inline unsigned int get_cyclecount(void) {
	unsigned int value;
	// Read CCNT Register
	asm volatile ("MRC p15, 0, %0, c9, c13, 0\t\n": "=r"(value));
	return value;
}

/*
 * periodically remove memory permissions of regions mapped to SoC modules.
 */
void my_timer_callback(unsigned long data) {
	int ret;
	u32 v;
	unsigned long long tmp_time;
	struct removed_base *p_removed_base;

	unsigned int timer_handler_start, timer_handler_end;

	timer_handler_start = get_cyclecount();

	list_for_each_entry(p_removed_base, &removed_base_list, node)
	{
		if (p_removed_base->accessed == true) {
			//module is busy, remove permissions
			p_removed_base->accessed = false;
			remove_permissions((unsigned long) (p_removed_base->base), true,
					NULL);
		} else {
			v = read_clkctrl_from_oh(p_removed_base->oh);
			v &= OMAP4430_MODULEMODE_MASK;
			v = v >> OMAP4430_MODULEMODE_SHIFT;
			if (v != 0) {
				//module is idle and is still enabled, disable it
				start_time = (unsigned long long) ktime_to_ns(ktime_get());
				_disable_module(p_removed_base->oh);
				/* toggle this line to enalbe/disalbe output */
				PRINTK0("disable %s\n", p_removed_base->oh->name);
			}
		}
	}
	ret = mod_timer(&my_timer, jiffies + msecs_to_jiffies(MONITOR_PERIOD));

	timer_handler_end = get_cyclecount();
	PRINTK3("cycles = %u, %s\n", timer_handler_end-timer_handler_start);
	if (ret)
		printk("Error in mod_timer\n");
}

/*
 * Interface to read how much time a SoC module is disabled
 */
ssize_t time_read(struct file *file, char __user *buf, size_t size, loff_t *ppos) {
	static int finished = 0;
	if (finished) {
		//printk(KERN_INFO "procfs_read: END\n");
		finished = 0;
		return 0;
	}
	finished = 1;

	memset(cmd_buffer, 0, sizeof(cmd_buffer));

	sprintf(cmd_buffer + strlen(cmd_buffer),
			"\ncentral put time is %llu\n",
			central_elapsed_time);
	copy_to_user(buf, cmd_buffer, strlen(cmd_buffer));
	return strlen(cmd_buffer);
}

/*
 * Show all SoC modules controlled by the Central PM Agent
 */
ssize_t cmd_read(struct file *file, char __user *buf, size_t size, loff_t *ppos) {
	static int finished = 0;
	struct removed_base *p_removed_base;

	if (finished) {
		//printk(KERN_INFO "procfs_read: END\n");
		finished = 0;
		return 0;
	}
	finished = 1;

	memset(cmd_buffer, 0, sizeof(cmd_buffer));

	list_for_each_entry(p_removed_base, &removed_base_list, node)
	{
		sprintf(cmd_buffer + strlen(cmd_buffer), "%s is in the list\n",
				p_removed_base->oh->name);
	}
	copy_to_user(buf, cmd_buffer, strlen(cmd_buffer));
	return strlen(cmd_buffer);
}

/*
 * Set a SoC module to be controlled by Central PM Agent
 */
int cmd_write(struct file *file, const char *buffer, unsigned long count,
		void *data) {
	int tmp;
	cmd_buffer_size = count;
	if (cmd_buffer_size > CMD_PROCFS_MAX_SIZE) {
		cmd_buffer_size = CMD_PROCFS_MAX_SIZE;
	}
	memset(cmd_buffer, 0, sizeof(cmd_buffer));
	if (copy_from_user(cmd_buffer, buffer, cmd_buffer_size)) {
		return -EFAULT;
	}
	cmd_buffer[strlen(cmd_buffer) - 1] = '\0';
	if (!strcmp(cmd_buffer, "gpio"))
		setup_gpio_base_protection(-1);
	else if (strstr(cmd_buffer, "gpio"))
		setup_gpio_base_protection(cmd_buffer[4] - '0');
	else if (!strcmp(cmd_buffer, "i2c"))
		setup_i2c_base_protection(-1);
	else if (!strcmp(cmd_buffer, "mmc"))
		setup_mmc_base_protection(-1);
	else if (strstr(cmd_buffer, "mmc"))
		setup_mmc_base_protection(cmd_buffer[3] - '0');
	else if (!strcmp(cmd_buffer, "dispc"))
		setup_dss_base_protection();
	else if (!strcmp(cmd_buffer, "measure page fault latency")) {
		for (tmp = 0; tmp < 100; tmp++) {
			measure_page_fault();
		}
	}
	return cmd_buffer_size;
}

/*
 * Page fault will be redirected to this function, by __do_kernel_fault()
 * in arch/arm/mm/fault.c
 */
int page_access_fix(unsigned long vaddr, unsigned int fsr, struct mm_struct *mm) {
	struct remapped_section *sec;
	pmd_t *pmdp;
	pte_t *ptep;
	bool found = false;
	struct removed_base *p_removed_base;
	unsigned long hpte;
	unsigned long flags;
	int r;

	if (mm == NULL) {
		printk(KERN_ALERT "CHAO: %s: mm=null! Should never occur!\n", __func__);
	}

	pmdp = pmd_offset(pgd_offset(mm, vaddr), vaddr);
	ptep = pte_offset_kernel(pmdp, vaddr);

	if (ptep == NULL) {
		printk("%s [2]: vaddr=%#X has ptep=null, so RETURN.\n", __FUNCTION__,
				vaddr);
		return VM_BAD_ACCESS;
	}

	/*
	 * Check if the page fault is caused by the memory access to the
	 * memory regions of a SoC module that is controlled by the
	 * Central PM Agent
	 */
	list_for_each_entry(p_removed_base, &removed_base_list, node)
	{
		if (p_removed_base->pfn == (*ptep) >> PAGE_SHIFT) {
			found = true;
			p_removed_base->accessed = true;
			/*
			 * These code profile how much time a module is disabled
			 */
			end_time = (unsigned long long) ktime_to_ns(ktime_get());
			if (start_time!=start_time_copy){
				if (central_elapsed_time == 0) {
					central_elapsed_time = 1;
					start_time_copy = start_time;
				}
				else {
					central_elapsed_time += (end_time-start_time);
					start_time_copy = start_time;
				}
			}
			/*
			 * Enable the module, and wait until it's ready
			 */
			_enable_module(p_removed_base->oh);
			r=_wait_target_ready(p_removed_base->oh);
			if (r){
				PRINTK0("omap_hwmod: %s: _wait_target_ready: %d\n",
						p_removed_base->oh->name, r);
			}
			break;
		}
	}

	if (found) {
		PRINTK3("%s [2]: vaddr=%#X is inside a removed pgtable, ptep=%#X."
				"(*ptep) >> PAGE_SHIFT = %#X\n",
				__FUNCTION__, vaddr, ptep, (*ptep) >> PAGE_SHIFT);
	} else {
		PRINTK3("%s [2]: vaddr=%#X is NOT inside a removed pgtable."
				"(*ptep) >> PAGE_SHIFT = %#X\n",
				__FUNCTION__, vaddr, (*ptep) >> PAGE_SHIFT);
		return VM_BAD_ACCESS;
	}

	/*
	 * Clean the cache that store this page talbe entry
	 * It may be unnecessary. Cache should only be flushed after modifying
	 * the page table
	 */
	clean_dcache_area((void *)(hw_pte(ptep)), sizeof(pte_t));

	PRINTK3(
			"%s [3]: PTE=%#X, write=%s exec=%s present=%s dirty=%s young=%s special=%s\n",
			__FUNCTION__, *ptep, pte_write(*ptep) ? "true" : "false",
					pte_exec(*ptep) ? "true" : "false",
							pte_present(*ptep) ? "true" : "false",
									pte_dirty(*ptep) ? "true" : "false",
											pte_young(*ptep) ? "true" : "false",
													pte_special(*ptep) ? "true" : "false");

	PRINTK3("%s [4]: fsr showed it was a %s attempt.\n", __FUNCTION__,
			fsr & FSR_WRITE ? "write" : "read");


	/* pte was not present, indicating NO write/read permissions */
	if (!pte_present(*ptep)) {
		set_pte_ext(ptep, pte_mkpresent(*ptep), 0);
		PRINTK3("%s [5]: PTE wasn't present, now setting it to be present.\n",
				__FUNCTION__);

		if (fsr & FSR_WRITE) {
			set_pte_ext(ptep, pte_mkwrite(*ptep), 0);
			PRINTK3(
					"%s [6]: was a write attempt, granting read and write permissions.\n",
					__FUNCTION__);
		} else {
			set_pte_ext(ptep, pte_wrprotect(*ptep), 0);
			PRINTK3(
					"%s [6]: was a read attempt, granting read-only permissions.\n",
					__FUNCTION__);
		}
	}
	/* pte was present, possibly indicating no write (read-only) permissions */
	else {
		PRINTK3("%s [5]: PTE was already present.\n", __FUNCTION__);
		if (fsr & FSR_WRITE) {
			set_pte_ext(ptep, pte_mkwrite(*ptep), 0);
			PRINTK3(
					"%s [6]: was a write attempt, granting read and write permission.\n",
					__FUNCTION__);
		} else {
			PRINTK3("%s [6]: PTE was present and faulted on a read access...\n",
					__FUNCTION__);
		}
	}

	/* Clean the cache that store this page talbe entry */
	clean_dcache_area((void *)(hw_pte(ptep)), sizeof(pte_t));
	hpte = readl(hw_pte(ptep));
	flush_tlb_kernel_page(vaddr & PAGE_MASK);


	list_for_each_entry(p_removed_base, &removed_base_list, node)
	{

		p_removed_base->accessed = true;
		break;
	}
	PRINTK3("%s [end] final PTE=%#X\n", __FUNCTION__, *ptep);

	return 0;
}
EXPORT_SYMBOL(page_access_fix);

/*
 * If 'all' is true, then remove both read and write permissions
 * otherwise, remove only write permissions
 */
void remove_permissions(unsigned long vaddr, bool all,
		struct removed_base *removed_base_inst) {
	int ret;
	pgd_t * pgdp;
	unsigned long hpte;
	unsigned long flags;
	u32 new_ptep;

	PRINTK2("%s, [beg] vaddr = %#X\n", __func__, vaddr);
	/*
	 * Before using the init_mm, this code tries to use a bunch of other
	 * "mm", which are unnecessary, because we only modify kernel pages.
	 */
	if (current->mm != NULL) {
		pgdp = pgd_offset(current->mm, vaddr);
		PRINTK3("%s, current->mm->pgd = %#X\n", __func__, current->mm->pgd);
	} else {
		PRINTK2("current->mm is null\n");
		if (current->active_mm != NULL) {
			pgdp = pgd_offset(current->active_mm, vaddr);
			PRINTK3("%s, current->active_mm->pgd = %#X\n", __func__,
					current->active_mm->pgd);
		} else {
			PRINTK3("both mm and active_mm are null!!!\n");
			PRINTK3("use init_mm\n");
			pgdp = pgd_offset(&init_mm, vaddr);
			PRINTK3("%s, init_mm.pgd = %#X\n", __func__, init_mm.pgd);
		}
	}
	PRINTK2("%s, pgdp = %#X\n", __func__, pgdp);
	pud_t * pudp = pud_offset(pgdp, vaddr);
	PRINTK2("%s, pudp = %#X\n", __func__, pudp);
	pmd_t * pmdp = pmd_offset(pudp, vaddr);
	PRINTK2("%s, pmd = %#X\n", __func__, *pmdp);
	PRINTK2("%s, pmdp = %#X\n", __func__, pmdp);
	PRINTK2("%s, pmd type=%08x\n", __func__, *pmdp & PMD_TYPE_MASK);
	if ((*pmdp & PMD_TYPE_MASK) == PMD_TYPE_SECT) {
		printk("%s [4]: pmd points to a %s, no PTE.\n", __FUNCTION__,
				((*pmdp & PMD_SECT_SUPER) ? "Supersection" : "Section"));
		return;
	}

	pte_t *ptep = pte_offset_kernel(pmdp, vaddr);

	hpte = readl(hw_pte(ptep));

	if (removed_base_inst != NULL) {
		removed_base_inst->ptep = ptep;
		removed_base_inst->pfn = (*ptep) >> PAGE_SHIFT;
	}
	PRINTK3("%s [beg]: PTEP=%#X\n", ptep);
	PRINTK3("%s [beg]: PTE=%#X, hwpte=0x%08x, read=%s write=%s\n", __FUNCTION__,
			*ptep, *(pte_t *)(hw_pte(ptep)),
			pte_present(*ptep) ? "true" : "false",
					pte_write(*ptep) ? "true" : "false");


	if (all) {
		set_pte_ext(ptep, pte_mknotpresent(*ptep), 0);
	}
	set_pte_ext(ptep, pte_exprotect(*ptep), 0); // this is bit 9 of the PTE, which vmalloc'd PTEs have set to 1
	set_pte_ext(ptep, pte_wrprotect(*ptep), 0);

	/*
	 * Flush the cache for the page talbe entry and flush the tlb
	 */
	clean_dcache_area((void *)(hw_pte(ptep)), sizeof(pte_t));
	flush_tlb_kernel_page(vaddr & PAGE_MASK);

	PRINTK2("%s [end]: PTE=%#X, hwpte=0x%08x, read=%s write=%s\n", __FUNCTION__,
			*ptep, *(pte_t *)(hw_pte(ptep)),
			pte_present(*ptep) ? "true" : "false",
					pte_write(*ptep) ? "true" : "false");
	return;
}

static const struct file_operations cmd_proc_fops = { .owner = THIS_MODULE,
		.read = cmd_read, .write = cmd_write, };
static const struct file_operations time_proc_fops = { .owner = THIS_MODULE,
		.read = time_read, };

static int __init init_pgtest(void) {
	char *c, *ptr, letter;
	PRINTK3("%s [0]: loaded module.\n", __FUNCTION__);
	/* create the /proc file */
	cmd_procfs = proc_create(CMD_PROCFS_NAME, S_IFREG | S_IRUGO, NULL,
			&cmd_proc_fops);
	if (cmd_procfs == NULL) {
		remove_proc_entry(CMD_PROCFS_NAME, NULL);
		printk(KERN_ALERT "Error: Could not initialize /proc/%s\n",
				CMD_PROCFS_NAME);
		return -ENOMEM;
	}

	time_procfs = proc_create(TIME_PROCFS_NAME, S_IFREG | S_IRUGO, NULL,
				&time_proc_fops);
	if (time_procfs == NULL) {
		remove_proc_entry(TIME_PROCFS_NAME, NULL);
		printk(KERN_ALERT "Error: Could not initialize /proc/%s\n",
				BUG_PROCFS_NAME);
		return -ENOMEM;
	}

	init_perfcounters(1, 0);
	rmpgtbl_ops = &ops;
	INIT_LIST_HEAD(&removed_base_list);
	setup_timer(&my_timer, my_timer_callback, 0);
	return 0;
}

static void __exit exit_pgtest(void) {
	int ret;
	if (cmd_procfs != NULL) {
		PRINTK3("CHAO: remove the cmd interface\n");
		remove_proc_entry(CMD_PROCFS_NAME, NULL);
	}
	if (time_procfs != NULL) {
		PRINTK3("CHAO: remove the time interface\n");
		remove_proc_entry(TIME_PROCFS_NAME, NULL);
	}
	ret = del_timer(&my_timer);
	if (ret)
		printk("The timer is still in use...\n");
	printk("CHAO: printk always show up!\n");

	PRINTK3("\n\n%s [0]: Goodbye world!\n", __FUNCTION__);
}

/*
 * If bank_id==-1, then setup protection for all banks
 * gpio0-5
 */
void setup_gpio_base_protection(int bank_id) {
	struct gpio_bank *bank;
	struct removed_base *removed_base_inst;
	struct omap_hwmod *oh;
	void __iomem *base;
	u32 value;
	int i;
	int ret;
	if (bank_id > 5 || bank_id < -1) {
		printk("CHAO: %s: invaild bank_id %d!\n", __func__, bank_id);
		return;
	}
	if (bank_id == -1) {
		list_for_each_entry(bank, pomap_gpio_list, node) {
			PRINTK3("%s, bank->base = 0x%08x\n", __func__, bank->base);
			removed_base_inst = kzalloc(sizeof(struct removed_base),
					GFP_KERNEL);
			removed_base_inst->base = bank->base;
			removed_base_inst->accessed = false;
			removed_base_inst->ptep = NULL;
			removed_base_inst->oh = to_omap_device(
					to_platform_device(bank->dev))->hwmods[0];
			removed_base_inst->catched_times = 0;
			removed_base_inst->abs_catched_times = 0;
			PRINTK3(KERN_ALERT "CHAO: oh name = %s\n",
					removed_base_inst->oh->name);
			list_add_tail(&removed_base_inst->node, &removed_base_list);
			remove_permissions((unsigned long) (bank->base), true,
					removed_base_inst);
		}
	} else {
		bank = container_of(pomap_gpio_list, struct gpio_bank, node);
		for (i = 0; i <= bank_id; i++) {
			bank = container_of(bank->node.next, struct gpio_bank, node);
		}
		PRINTK3("%s, bank->base = 0x%08x\n", __func__, bank->base);
		removed_base_inst = kzalloc(sizeof(struct removed_base), GFP_KERNEL);
		removed_base_inst->base = bank->base;
		removed_base_inst->accessed = false;
		removed_base_inst->ptep = NULL;
		removed_base_inst->oh =
				to_omap_device(to_platform_device(bank->dev))->hwmods[0];
		removed_base_inst->catched_times = 0;
		removed_base_inst->abs_catched_times = 0;
		list_add_tail(&removed_base_inst->node, &removed_base_list);
		remove_permissions((unsigned long) (bank->base), true,
				removed_base_inst);

	}
	ret = mod_timer(&my_timer, jiffies + msecs_to_jiffies(MONITOR_PERIOD));
	if (ret)
		printk("Error in mod_timer\n");
}

/*
 * If i2c_id==-1, then setup protection for all i2c
 * i2c1-4
 */
void setup_i2c_base_protection(int i2c_id) {
	struct omap_i2c_dev *dev;
	struct removed_base *removed_base_inst;
	void __iomem *base;
	u32 value;
	int i;
	int ret;
	if (i2c_id > 4 || (i2c_id < 1 && i2c_id != -1)) {
		printk("CHAO: %s: invaild i2c_id %d!\n", __func__, i2c_id);
		return;
	}
	if (i2c_id == -1) {
		list_for_each_entry(dev, pomap_i2c_list, node) {
			PRINTK3("%s, dev->base = 0x%08x\n", __func__, dev->base);
			removed_base_inst = kzalloc(sizeof(struct removed_base),
					GFP_KERNEL);
			removed_base_inst->base = dev->base;
			removed_base_inst->accessed = false;
			removed_base_inst->ptep = NULL;
			removed_base_inst->oh =
					to_omap_device(to_platform_device(dev->dev))->hwmods[0];
			removed_base_inst->catched_times = 0;
			removed_base_inst->abs_catched_times = 0;
			removed_base_inst->abs_catched_times = 0;
			list_add_tail(&removed_base_inst->node, &removed_base_list);
			remove_permissions((unsigned long) (dev->base), true,
					removed_base_inst);
		}
	} else {
		dev = container_of(pomap_i2c_list, struct omap_i2c_dev, node);
		for (i = 0; i < i2c_id; i++) {
			dev = container_of(dev->node.next, struct omap_i2c_dev, node);
		}
		PRINTK3("%s, dev->base = 0x%08x\n", __func__, dev->base);
		removed_base_inst = kzalloc(sizeof(struct removed_base), GFP_KERNEL);
		removed_base_inst->base = dev->base;
		removed_base_inst->accessed = false;
		removed_base_inst->ptep = NULL;
		removed_base_inst->oh =
				to_omap_device(to_platform_device(dev->dev))->hwmods[0];
		removed_base_inst->catched_times = 0;
		removed_base_inst->abs_catched_times = 0;
		list_add_tail(&removed_base_inst->node, &removed_base_list);
		remove_permissions((unsigned long) (dev->base), true,
				removed_base_inst);
	}
	ret = mod_timer(&my_timer, jiffies + msecs_to_jiffies(MONITOR_PERIOD));
	if (ret)
		printk("Error in mod_timer\n");
}

/*
 * If mmc_id==-1, then setup protection for all i2c
 * mmc1-5
 * CHAO: omap_mmc_probe: SZ_4k
 */
void setup_mmc_base_protection(int mmc_id) {
	struct omap_hsmmc_host *host;
	struct removed_base *removed_base_inst;
	void __iomem *base;
	u32 value;
	int i;
	int ret;
	if (mmc_id > 5 || (mmc_id < 1 && mmc_id != -1)) {
		printk("CHAO: %s: invaild mmc_id %d!\n", __func__, mmc_id);
		return;
	}
	if (mmc_id == -1) {
		/* disable the driver's original PM in omap_hsmmc.c */
		mmc_pm = 0;
		list_for_each_entry(host, pomap_mmc_list, node)
										{
			PRINTK3("%s, host->base = 0x%08x\n", __func__, host->base);
			removed_base_inst = kzalloc(sizeof(struct removed_base),
					GFP_KERNEL);
			removed_base_inst->base = host->base;
			removed_base_inst->accessed = false;
			removed_base_inst->ptep = NULL;
			removed_base_inst->oh = to_omap_device(
					to_platform_device(host->dev))->hwmods[0];
			PRINTK3("%s, oh->name = %s\n", __func__,
					removed_base_inst->oh->name);
			removed_base_inst->catched_times = 0;
			removed_base_inst->abs_catched_times = 0;
			list_add_tail(&removed_base_inst->node, &removed_base_list);
			remove_permissions((unsigned long) (host->base), true,
					removed_base_inst);
										}
	} else {
		/* disable the driver's original PM in omap_hsmmc.c */
		mmc_pm = 0;
		host = container_of(pomap_mmc_list, struct omap_hsmmc_host, node);
		for (i = 0; i < mmc_id; i++) {
			host = container_of(host->node.next, struct omap_hsmmc_host, node);
		}
		PRINTK3("%s, host->base = 0x%08x\n", __func__, host->base);
		removed_base_inst = kzalloc(sizeof(struct removed_base), GFP_KERNEL);
		removed_base_inst->base = host->base;
		removed_base_inst->accessed = false;
		removed_base_inst->ptep = NULL;
		removed_base_inst->oh =
				to_omap_device(to_platform_device(host->dev))->hwmods[0];
		PRINTK3("%s, oh->name = %s\n", __func__, removed_base_inst->oh->name);
		removed_base_inst->catched_times = 0;
		removed_base_inst->abs_catched_times = 0;
		list_add_tail(&removed_base_inst->node, &removed_base_list);
		remove_permissions((unsigned long) (host->base), true,
				removed_base_inst);
	}
	ret = mod_timer(&my_timer, jiffies + msecs_to_jiffies(MONITOR_PERIOD));
	if (ret)
		printk("Error in mod_timer\n");
}

void setup_dss_base_protection(void) {
	struct struct_dispc *dispc = (struct struct_dispc *)pomap_dispc;
	struct removed_base *removed_base_inst;
	void __iomem *base;
	u32 value;
	int i;
	int ret;

	PRINTK0("%s, host->base = 0x%08x\n", __func__, dispc->base);
	removed_base_inst = kzalloc(sizeof(struct removed_base),
			GFP_KERNEL);
	removed_base_inst->base = dispc->base;
	removed_base_inst->accessed = false;
	removed_base_inst->ptep = NULL;
	removed_base_inst->oh = to_omap_device(
			dispc->pdev)->hwmods[0];
	PRINTK0("%s, oh->name = %s\n", __func__,
			removed_base_inst->oh->name);
	removed_base_inst->catched_times = 0;
	removed_base_inst->abs_catched_times = 0;
	list_add_tail(&removed_base_inst->node, &removed_base_list);
	remove_permissions((unsigned long) (dispc->base), true,
			removed_base_inst);
	ret = mod_timer(&my_timer, jiffies + msecs_to_jiffies(MONITOR_PERIOD));
	if (ret)
		printk("Error in mod_timer\n");
}

void measure_page_fault(void) {
	struct gpio_bank *bank;
	struct removed_base *removed_base_inst;
	struct omap_hwmod *oh;
	unsigned int start_time;
	unsigned int end_time;
	unsigned int tmp1, tmp2;
	void __iomem *base;
	u32 value1, value2;
	int i;
	int ret;
	unsigned long local_page_fix_count;
	/* Use gpio_3 for this experiment */
	bank = container_of(pomap_gpio_list, struct gpio_bank, node);
	bank = container_of(bank->node.next, struct gpio_bank, node);
	bank = container_of(bank->node.next, struct gpio_bank, node);
	bank = container_of(bank->node.next, struct gpio_bank, node);
	PRINTK3("%s, bank->base = 0x%08x\n", __func__, bank->base);
	removed_base_inst = kzalloc(sizeof(struct removed_base), GFP_KERNEL);
	removed_base_inst->base = bank->base;
	removed_base_inst->accessed = false;
	removed_base_inst->ptep = NULL;
	removed_base_inst->oh =
			to_omap_device(to_platform_device(bank->dev))->hwmods[0];
	PRINTK2("CHAO: oh->name=%s\n", removed_base_inst->oh->name);
	removed_base_inst->catched_times = 0;
	removed_base_inst->abs_catched_times = 0;
	value1 = __raw_readl(bank->base);
	list_add_tail(&removed_base_inst->node, &removed_base_list);
	remove_permissions((unsigned long) (bank->base), true, removed_base_inst);

	local_page_fix_count = page_fix_access_count;
	PRINTK2("CHAO: before access, count=%lu\n", page_fix_access_count);
	start_time = get_cyclecount();
	value1 = __raw_readl(bank->base);
	end_time = get_cyclecount();
	PRINTK2("CHAO: after access, count=%lu\n", page_fix_access_count);
	tmp1 = end_time - start_time;
	start_time = get_cyclecount();
	value2 = __raw_readl(bank->base);
	end_time = get_cyclecount();
	tmp2 = end_time - start_time;
	PRINTK0("CHAO: first access: %u, second access: %u\n, value1=0x%08x, value2=0x%08x\n",
			tmp1, tmp2, value1, value2);
}

void test_get_cycle(void) {
	unsigned int start_time;
	unsigned int end_time;
	unsigned int tmp1, tmp2;
	start_time = get_cyclecount();
	get_cyclecount();
	end_time = get_cyclecount();
	tmp2 = end_time - start_time;
	printk("CHAO: %s: %u\n", __func__, tmp2);
}
module_init(init_pgtest);
module_exit(exit_pgtest);
